/*
    id3extract.c (MP3 id3 extractor) is part of
    flc 1999-2002 by NoisyB (noisyb@gmx.net)
    flc lists information about the FILEs (the current directory by default)
    But most important it shows the FILE_ID.DIZ of every file (if present)
    Very useful for FTP Admins or people who have a lot to do with FILE_ID.DIZ
    Copyright (C) 1999-2002 by NoisyB (noisyb@gmx.net)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/
#include <stdio.h>
#include <string.h>
#include <time.h>
#include "config.h"
#include "misc.h"
#include "id3extract.h"

#define EMPTY_TO_UNKNOWN_S(str) (str[0] ? (char *)str : "?")

static int id3extract (const char *mp3file);

int
main (int argc, char *argv[])
{
  if (argc < 2)
    {
      printf ("USAGE: %s MP3FILE\n\n"
              "extracts ID3v1 tag from any MP3FILE\n\n", argv[0]);

      return -1;
    }
  id3extract (argv[1]);

  return 0;
}

int
id3extract (const char *mp3file)
{
  id3v1_tag_info_t id3v1;
//  id3v2_tag_info_t id3v2; // id3v2 sucks
  int result = 0;
  FILE *fh;
#define ID3V1_GENRE_MAX 148
  static const char *id3v1_genre[ID3V1_GENRE_MAX] =
    {
  "(00)  Blues",
  "(01)  Classic Rock",
  "(02)  Country",
  "(03)  Dance",
  "(04)  Disco",
  "(05)  Funk",
  "(06)  Grunge",
  "(07)  Hip-Hop",
  "(08)  Jazz",
  "(09)  Metal",
  "(10)  New Age",
  "(11)  Oldies",
  "(12)  Other",
  "(13)  Pop",
  "(14)  R&B",
  "(15)  Rap",
  "(16)  Reggae",
  "(17)  Rock",
  "(18)  Techno",
  "(19)  Industrial",
  "(20)  Alternative",
  "(21)  Ska",
  "(22)  Death Metal",
  "(23)  Pranks",
  "(24)  Soundtrack",
  "(25)  Euro-Techno",
  "(26)  Ambient",
  "(27)  Trip-Hop",
  "(28)  Vocal",
  "(29)  Jazz+Funk",
  "(30)  Fusion",
  "(31)  Trance",
  "(32)  Classical",
  "(33)  Instrumental",
  "(34)  Acid",
  "(35)  House",
  "(36)  Game",
  "(37)  Sound Clip",
  "(38)  Gospel",
  "(39)  Noise",
  "(40)  Alternative Rock",
  "(41)  Bass",
  "(43)  Punk",
  "(44)  Space",
  "(45)  Meditative",
  "(46)  Instrumental Pop",
  "(47)  Instrumental Rock",
  "(48)  Ethnic",
  "(49)  Gothic",
  "(50)  Darkwave",
  "(51)  Techno-Industrial",
  "(52)  Electronic",
  "(53)  Pop-Folk",
  "(54)  Eurodance",
  "(55)  Dream",
  "(56)  Southern Rock",
  "(57)  Comedy",
  "(58)  Cult",
  "(59)  Gangsta",
  "(60)  Top 40",
  "(61)  Christian Rap",
  "(62)  Pop/Funk",
  "(63)  Jungle",
  "(64)  Native US",
  "(65)  Cabaret",
  "(66)  New Wave",
  "(67)  Psychadelic",
  "(68)  Rave",
  "(69)  Showtunes",
  "(70)  Trailer",
  "(71)  Lo-Fi",
  "(72)  Tribal",
  "(73)  Acid Punk",
  "(74)  Acid Jazz",
  "(75)  Polka",
  "(76)  Retro",
  "(77)  Musical",
  "(78)  Rock & Roll",
  "(79)  Hard Rock",
  "(80)  Folk",
  "(81)  Folk-Rock",
  "(82)  National Folk",
  "(83)  Swing",
  "(84)  Fast Fusion",
  "(85)  Bebob",
  "(86)  Latin",
  "(87)  Revival",
  "(88)  Celtic",
  "(89)  Bluegrass",
  "(90)  Avantgarde",
  "(91)  Gothic Rock",
  "(92)  Progressive Rock",
  "(93)  Psychedelic Rock",
  "(94)  Symphonic Rock",
  "(95)  Slow Rock",
  "(96)  Big Band",
  "(97)  Chorus",
  "(98)  Easy Listening",
  "(99)  Acoustic",
  "(100) Humour",
  "(101) Speech",
  "(102) Chanson",
  "(103) Opera",
  "(104) Chamber Music",
  "(105) Sonata",
  "(106) Symphony",
  "(107) Booty Bass",
  "(108) Primus",
  "(109) Porn Groove",
  "(110) Satire",
  "(111) Slow Jam",
  "(112) Club",
  "(113) Tango",
  "(114) Samba",
  "(115) Folklore",
  "(116) Ballad",
  "(117) Power Ballad",
  "(118) Rhytmic Soul",
  "(119) Freestyle",
  "(120) Duet",
  "(121) Punk Rock",
  "(122) Drum Solo",
  "(123) Acapella",
  "(124) Euro-House",
  "(125) Dance Hall",
  "(126) Goa",
  "(127) Drum & Bass",
  "(128) Club-House",
  "(129) Hardcore",
  "(130) Terror",
  "(131) Indie",
  "(132) BritPop",
  "(133) Negerpunk",
  "(134) Polsk Punk",
  "(135) Beat",
  "(136) Christian Gangsta",
  "(137) Heavy Metal",
  "(138) Black Metal",
  "(139) Crossover",
  "(140) Contemporary C",
  "(141) Christian Rock",
  "(142) Merengue",
  "(143) Salsa",
  "(144) Thrash Metal",
  "(145) Anime",
  "(146) JPop",
  "(147) SynthPop",
  "(148) ?"
    };

  result = quick_fread (&id3v1, file_size (mp3file) - sizeof (id3v1_tag_info_t), sizeof (id3v1_tag_info_t), mp3file);

  if (id3v1.magic[0] != 'T' ||
      id3v1.magic[1] != 'A' ||
      id3v1.magic[2] != 'G') return -1;

  if (!(fh = fopen ("FILE_ID.DIZ", "wb"))) return -1;

  fprintf (fh, "Title  : %-30.30s\nArtist : %-30.30s\nAlbum  : %-30.30s\n"
//    "Track  : %d\n"
    "Year   : %-4.4s\nComment: %-30.30s\nGenre  : %s\n ",
    EMPTY_TO_UNKNOWN_S (id3v1.title),
    EMPTY_TO_UNKNOWN_S (id3v1.artist),
    EMPTY_TO_UNKNOWN_S (id3v1.album),
//    (id3v1.comment[29] ? id3v1.comment[29] : '?'),
    EMPTY_TO_UNKNOWN_S (id3v1.year),
    EMPTY_TO_UNKNOWN_S (id3v1.comment),
    &id3v1_genre[MIN (id3v1.genre, ID3V1_GENRE_MAX - 1)][6]
    );            
  fclose (fh);

  return result;
}
